// -----------------------------------------------------------------------------
// Copyright (c) 2025 Proton AG
//
// This file is part of ProtonVPN.
//
// ProtonVPN is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// ProtonVPN is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------

pub mod network_manager;
pub mod protun;

use super::service::Service;
use std::sync::Arc;
use tokio::sync::{Notify, RwLock};

pub fn new_interfaces(shutdown: Arc<Notify>) -> (network_manager::NetworkManager, protun::Protun) {
    let state = Arc::new(RwLock::new(Service::default()));
    (network_manager::NetworkManager::new(state.clone(), shutdown), protun::Protun::new(state))
}
